// JOTM/XAPool Test Class
// This is to test transactions and connection
// Written By: Andy Zeneski (jaz@ofbiz.org)

import java.io.StringWriter;
import java.io.Writer;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.Date;
import java.util.Properties;
import java.util.Random;

import javax.transaction.HeuristicMixedException;
import javax.transaction.HeuristicRollbackException;
import javax.transaction.InvalidTransactionException;
import javax.transaction.NotSupportedException;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.UserTransaction;

import org.apache.log4j.Logger;
import org.apache.log4j.Priority;
import org.enhydra.jdbc.pool.StandardXAPoolDataSource;
import org.enhydra.jdbc.standard.StandardXADataSource;
import org.objectweb.jotm.Jotm;
import org.objectweb.transaction.jta.TMService;

public class JotmXaPoolTest {

    protected static Logger logger = Logger.getLogger("jotm.test");
    protected static String tableName = "jotm_xapool_test"; 
    
    protected String createColumns = null;
    protected int counter = 0;
    protected TMService jotm = null;
    protected StandardXAPoolDataSource pool = null;
    
    public JotmXaPoolTest() throws Exception {        
        // start JOTM
        jotm = new Jotm(true, false);
        logger.info("Started JOTM...");
        
        // need a classloader
        ClassLoader cl = Thread.currentThread().getContextClassLoader();
        
        // load the database properties
        Properties props = new Properties();       
        props.load(cl.getResourceAsStream("config/datasource.properties"));
        
        // set the sql for the create statement
        createColumns = props.getProperty("sql.create.columns");
        
        // get an instance of the datasource wrapper
        String defaultWrapper = "org.enhydra.jdbc.standard.StandardXADataSource";
        String wrapperName = props.getProperty("jdbc.wrapper", defaultWrapper);
        
        Class c = cl.loadClass(wrapperName);
        Object o = c.newInstance();
                
        // create the datasource
        StandardXADataSource ds = (StandardXADataSource) o;
        ds.setTransactionManager(jotm.getTransactionManager());
                
        ds.setDriverName(props.getProperty("jdbc.driver"));
        ds.setUrl(props.getProperty("jdbc.url"));
        ds.setUser(props.getProperty("jdbc.login"));
        ds.setPassword(props.getProperty("jdbc.password", ""));
        ds.setDescription(props.getProperty("jdbc.description", "NA"));
        
        // get the isolation level (default is READ_UNCOMITTED
        int isoLevel = 1;
        try {
            isoLevel = Integer.parseInt(props.getProperty("jdbc.isolation.level", "1"));
        } catch (NumberFormatException e) {
            logger.error("Problems parsing the isolation level (should be a number) using READ_UNCOMMITED");
            isoLevel = 1;
        }        
        ds.setTransactionIsolation(isoLevel);
                                
        // create the pool
        pool = new StandardXAPoolDataSource();
        pool.setDataSource(ds);
        pool.setDescription(ds.getDescription());
        pool.setUser(ds.getUser());
        pool.setPassword(ds.getPassword());
        pool.setTransactionManager(jotm.getTransactionManager());
              
        logger.info("Created pool...");        
    }    
    
    public void dropTest() throws SQLException {        
        Connection con = pool.getConnection();
        Statement s = con.createStatement();
        s.executeUpdate("DROP TABLE " + tableName);
        s.close();
        con.close();
        logger.info("Table '" + tableName + "' dropped.");        
    }
    
    public void createTest() throws SQLException {                
        // get a connection
        Connection con = pool.getConnection();                
        String sql = "CREATE TABLE " + tableName + " " + this.createColumns;
                
        Statement s = con.createStatement();
        s.executeUpdate(sql);
        s.close();                               
        con.close();
        logger.info("Table '" + tableName + "' created.");                    
    }
    
    /** Tests a loop of inserts using different connections, same transaction */
    public void insertTest() {
        // start the transaction
        UserTransaction trans = jotm.getUserTransaction();
        try {
            trans.begin();
        } catch (NotSupportedException e1) {
            logger.error("", e1);                       
        } catch (SystemException e1) {
            logger.error("", e1);                        
        }
        
        logger.info("Beginning multiple insert/connection single transaction...");
        for (int i = 0; i < 500; i++) {            
            Connection con = null;
            try {
                con = pool.getConnection();
            } catch (SQLException e) {            
                logger.error("Problems getting new connection - Test Failed!", e);
                return;
            }    
            if (con == null) {
                logger.error("Pool returned null connection with no exception - Test Failed!");
                return;    
            }
            
            logger.debug("Got connection.. inserting #" + i);
            try {            
                String sql = "INSERT INTO " + tableName + " VALUES(?,?,?,?)";
                PreparedStatement ps = con.prepareStatement(sql);
                ps.setInt(1, ++counter);
                ps.setString(2, "insTest" + i);
                ps.setString(3, "Insert Test");
                ps.setTimestamp(4, new Timestamp(new Date().getTime()));
                ps.executeUpdate();
                ps.close();                
            } catch (Exception e) {
                e.printStackTrace();
                try {
                    trans.setRollbackOnly();
                } catch (IllegalStateException e2) {                    
                    logger.error("", e2);
                } catch (SystemException e2) {                    
                    logger.error("", e2);
                }
            } finally {                
                try {
                    con.close();
                } catch (SQLException e2) {                    
                    logger.error("", e2);
                }
            }
        }
        try {        
            trans.commit();
        } catch (Exception e) {
            logger.error("", e);
        }
    }
    
    public void loopTest() {
        // start the transaction
        UserTransaction trans = jotm.getUserTransaction();
        try {
            trans.begin();            
        } catch (NotSupportedException e1) {            
            logger.error("", e1);
        } catch (SystemException e1) {            
            logger.error("", e1);
        }
                                
        logger.info("Beginning multiple insert single transaction/connection...");
        
        Connection con = null;
        try {
            con = pool.getConnection();
        } catch (SQLException e) {            
            logger.error("Problems getting new connection - Test Failed!", e);
            return;
        }    
        if (con == null) {
            logger.error("Pool returned null connection with no exception - Test Failed!");
            return;    
        }
        
        logger.debug("Got connection.. ");
        
        for (int i = 0; i < 500; i++) {                       
            logger.debug("Looping.. inserting #" + i);
            try {            
                String sql = "INSERT INTO " + tableName + " VALUES(?,?,?,?)";
                PreparedStatement ps = con.prepareStatement(sql);
                ps.setInt(1, ++counter);
                ps.setString(2, "loopTest" + i);
                ps.setString(3, "Loop Test");
                ps.setTimestamp(4, new Timestamp(new Date().getTime()));
                ps.executeUpdate();
                ps.close();                
            } catch (Exception e) {
                logger.error("", e);
                try {
                    trans.setRollbackOnly();
                } catch (IllegalStateException e2) {                    
                    logger.error("", e2);
                } catch (SystemException e2) {                    
                    logger.error("", e2);
                }
            }            
        }
        
        // close the connection
        try {
            con.close();
        } catch (SQLException e2) {            
            logger.error("", e2);
        }
        logger.debug("Closed connection..");
        
        // commit the transaction
        try {        
            trans.commit();            
        } catch (Exception e) {
            logger.error("", e);
        }
    }
    
    /** Tests multiple inserts; suspending before each one, insert in new trans, resuming to continue */
    public void suspendTest() {
        // start the transaction
        UserTransaction trans = jotm.getUserTransaction();
        TransactionManager tm = jotm.getTransactionManager();
        
        // begin the parent transaction
        try {
            trans.begin();
        } catch (NotSupportedException e1) {           
            logger.error("", e1);
        } catch (SystemException e1) {            
            logger.error("", e1);
        }
        
        logger.info("Beginning multiple insert/connection on suspend main transaction...");
        for (int i = 0; i < 500; i++) {            
            // suspend the main transaction                    
            Transaction transaction = null;            
            try {
                transaction = tm.suspend();
            } catch (SystemException e2) {                
                logger.error("", e2);
            }
            logger.debug("Suspended #" + i);
            
            // begin a new transaction
            try {
                trans.begin();
            } catch (NotSupportedException e3) {               
                logger.error("", e3);
            } catch (SystemException e3) {                
                logger.error("", e3);
            }
            
            // do some stuff in the new transaction
            Connection con1 = null;
            try {
                con1 = pool.getConnection();
            } catch (SQLException e) {               
                logger.error("Problems getting new (sub) connection - Test Failed!", e);
                return;
            }
            if (con1 == null) {
                logger.error("Pool returned null connection with no exception - Test Failed!");
                return;
            }
            try {
                String sql = "INSERT INTO " + tableName + " VALUES(?,?,?,?)";
                PreparedStatement ps = con1.prepareStatement(sql);
                ps.setInt(1, ++counter);
                ps.setString(2, "susTest" + i);
                ps.setString(3, "Suspend Test - Main Suspended");
                ps.setTimestamp(4, new Timestamp(new Date().getTime()));
                ps.executeUpdate();
                ps.close();                
            } catch (Exception e) {
                logger.error("", e);
                try {
                    trans.setRollbackOnly();
                } catch (IllegalStateException e4) {                    
                    logger.error("", e4);
                } catch (SystemException e4) {                    
                    logger.error("", e4);
                }
            } finally {                        
                try {
                    con1.close();
                } catch (SQLException e5) {                   
                    logger.error("", e5);
                }  
                
                // commit the new transaction              
                try {
                    trans.commit();
                } catch (SecurityException e4) {                   
                    logger.error("", e4);
                } catch (IllegalStateException e4) {                    
                    logger.error("", e4);
                } catch (RollbackException e4) {                    
                    logger.error("", e4);
                } catch (HeuristicMixedException e4) {                    
                    logger.error("", e4);
                } catch (HeuristicRollbackException e4) {                    
                    logger.error("", e4);
                } catch (SystemException e4) {                    
                    logger.error("", e4);
                }                                
            }
            
            // resume the main transaction
            try {
                tm.resume(transaction);
            } catch (InvalidTransactionException e4) {                
                logger.error("", e4);
            } catch (IllegalStateException e4) {                
                logger.error("", e4);
            } catch (SystemException e4) {                
                logger.error("", e4);
            }
            logger.debug("Resumed #" + i);
            
            // do some stuff in the main transaction
            Connection con2 = null;
            try {
                con2 = pool.getConnection();
            } catch (SQLException e) {
                logger.error("Problems getting new (main) connection - Test Failed!", e);
                return;              
            }          
            if (con2 == null) {
                logger.error("Pool returned null connection with no exception - Test Failed!");
                return;  
            }
            
            try {            
                String sql = "INSERT INTO " + tableName + " VALUES(?,?,?,?)";
                PreparedStatement ps = con2.prepareStatement(sql);
                ps.setInt(1, ++counter);
                ps.setString(2, "susTest" + i);
                ps.setString(3, "Suspend Test - Main");
                ps.setTimestamp(4, new Timestamp(new Date().getTime()));
                ps.executeUpdate();
                ps.close();
                logger.debug("Inserted main transaction.");
            } catch (Exception e) {
                logger.error("", e);
                try {
                    trans.setRollbackOnly();
                } catch (IllegalStateException e5) {                   
                    logger.error("", e5);
                } catch (SystemException e5) {                    
                    logger.error("", e5);
                }
            } finally {
                try {
                    con2.close();
                } catch (SQLException e5) {                    
                    logger.error("", e5);
                }
            }
        }
        
        // commit the main transaction - outside the loop
        try {
            trans.commit();
        } catch (SecurityException e) {            
            logger.error("", e);
        } catch (IllegalStateException e) {            
            logger.error("", e);
        } catch (RollbackException e) {            
            logger.error("", e);
        } catch (HeuristicMixedException e) {            
            logger.error("", e);
        } catch (HeuristicRollbackException e) {            
            logger.error("", e);
        } catch (SystemException e) {            
            logger.error("", e);
        }
    }
    
    public void rollbackOnlyTest() { 
        // start the transaction
        UserTransaction trans = jotm.getUserTransaction();
        try {
            trans.begin();
        } catch (NotSupportedException e1) {            
            logger.error("", e1);
        } catch (SystemException e1) {            
            logger.error("", e1);
        }
        
        logger.info("Beginning rollback test...");
        Random rand = new Random();
        int randomInt = rand.nextInt(9);
        
        for (int i = 0; i < 10; i++) {            
            Connection con = null;
            try {
                con = pool.getConnection();
            } catch (SQLException e) {                
                logger.error("Problems getting connection - rolling back now.", e);
                try {
                    trans.rollback();
                } catch (IllegalStateException e2) {                  
                    logger.error("", e2);
                } catch (SecurityException e2) {                    
                    logger.error("", e2);
                } catch (SystemException e2) {                    
                    logger.error("", e2);
                }
            }
            if (con == null) {
                logger.error("Pool returned a null connection w/ no exception! Test Failed!");
                return;
            }
            
            logger.debug("Got connection.. inserting #" + i);
            try {            
                String sql = "INSERT INTO " + tableName + " VALUES(?,?,?,?)";
                PreparedStatement ps = con.prepareStatement(sql);
                ps.setInt(1, ++counter);
                ps.setString(2, "rollTest" + i);
                ps.setString(3, "Rollback Test - This should not show in selectTest!");
                ps.setTimestamp(4, new Timestamp(new Date().getTime()));
                ps.executeUpdate();
                ps.close();                                  
            } catch (Exception e) {
                logger.error("", e);
                try {
                    trans.setRollbackOnly();
                } catch (IllegalStateException e2) {                   
                    logger.error("", e2);
                } catch (SystemException e2) {                    
                    logger.error("", e2);
                }
            } finally {                
                try {
                    con.close();
                } catch (SQLException e2) {                    
                    logger.error("", e2);
                }                
            }
            
            // will set rollback only on some random pass
            if (randomInt == i) {
                logger.info("Setting rollback only on pass #" + i);
                try {
                    trans.setRollbackOnly();
                } catch (IllegalStateException e2) {                   
                    logger.error("", e2);
                } catch (SystemException e2) {                    
                    logger.error("", e2);
                }
            }
            
        }
        try {        
            trans.commit();        
        } catch (RollbackException re) {
            // This SHOULD happen!
            logger.info("Commit failed (good), transaction rolled back by commit().");          
        } catch (SecurityException e) {
            logger.error("Commit failed; RollbackException not thrown!", e);                      
        } catch (IllegalStateException e) {            
            logger.error("Commit failed; RollbackException not thrown!", e);
        } catch (HeuristicMixedException e) {            
            logger.error("Commit failed; RollbackException not thrown!", e);
        } catch (HeuristicRollbackException e) {          
            logger.error("Commit failed; RollbackException not thrown!", e);
        } catch (SystemException e) {            
            logger.error("Commit failed; RollbackException not thrown!", e);
        }
    }            
    
    public void selectTest() throws SQLException {
        logger.info("Beginning select test.. We should have exactly 2000 records.");
        Connection con = pool.getConnection();
        Statement s = con.createStatement();        
        ResultSet res = s.executeQuery("SELECT * FROM " + tableName + " ORDER BY idx");
        int rowCount = 0;
        while (res.next()) {
            rowCount++;
            logger.debug(res.getString(1) + " : " + res.getString(2) + "[" + res.getString(3) + "] - " + res.getString(4));
        }
        res.close();
        con.close();
        logger.info("Total Rows: " + rowCount + " of 2000");
        if (rowCount == 2000) logger.info("Looks good...");        
    }
    
    public void close() {
        pool.shutdown(true);
        jotm.stop();
        logger.info("Shutdown pool and jotm.");                
    }
    
    public void runPreTests() {
        // try to drop the table
        try {       
            dropTest();
        } catch (SQLException e) {            
        }
        
        // try to create the table
        try {        
            createTest();
        } catch (SQLException e) {
            logger.error("Fatal SQL Error", e);
            this.close();
            return;
        }
    }
    
    public void runPostTests() {
        // show the results
        try {        
            selectTest();
        } catch (SQLException e) {
            logger.error("SQL Error", e);
        }
        
        // drop the table
        try {       
            dropTest();
        } catch (SQLException e) {
            logger.error("Fatal SQL Error", e);
            this.close(); 
            return;           
        }         
    }        
                    
    public void runAllTests() {       
        runPreTests();
        
        // test some basic inserts                              
        insertTest();        
        
        // test looping; same connection      
        loopTest();        
        
        // test some suspend/resume inserts            
        suspendTest();        
        
        // rollback only test       
        rollbackOnlyTest();        
        
        runPostTests();
    }
            
    public static void main(String[] args) throws Exception {           
        JotmXaPoolTest test = new JotmXaPoolTest();
        test.runAllTests();             
        System.exit(0);    
    }
}
